require "/scripts/vec2.lua"
require "/scripts/util.lua"

function init()

  self.specialLast = false
  self.active = false
  self.fireTimer = 0
  animator.rotateGroup("guns", 0, true)
  self.level = vehicle.configParameter("mechLevel", 6)
  self.groundFrames = 1
  
-- required for damaging
  self.maxHealth = vehicle.configParameter("maxHealth",1000)
  self.protection = vehicle.configParameter("protection",50)
  self.materialKind = vehicle.configParameter("materialKind","robotic")

  if storage.health == nil then
    local storedHP = vehicle.configParameter("startHealthFactor")

    if (storedHP == nil) then
        storage.health = self.maxHealth
    else
       storage.health = math.min(storedHP * self.maxHealth, self.maxHealth)
    end    
    self.warpPosition = mcontroller.position()
    animator.setAnimationState("movement", "warpIn")
  else
    animator.setAnimationState("movement", "idle")
  end
    
  --this comes in from the controller.
  self.ownerKey = vehicle.configParameter("ownerKey")
  vehicle.setPersistent(self.ownerKey)

   --setup the store functionality  
  message.setHandler("store",
                      function(_, _, ownerKey)

                        if (self.ownerKey and self.ownerKey == ownerKey and self.driver == nil and animator.animationState("movement")=="idle") then
                          self.warpPosition = mcontroller.position()
                          animator.setAnimationState("movement", "warpOut")
--                          vehicle.destroy()
                          return {storable = true, healthFactor = storage.health / self.maxHealth}
                        else
                          return {storable = false, healthFactor = storage.health / self.maxHealth}
                        end
                      end
                    )


end

--------------------------------------------------------------------------------

function update()

  local warpAni = animator.animationState("movement")
  if warpAni == "warpOutEnd" then
    vehicle.destroy()
    return
  elseif warpAni == "warpIn" or warpAni == "warpOut" then
    world.debugText("warping",mcontroller.position(),"red")

    --lock it solid whilst spawning/despawning
--    mcontroller.setPosition(self.warpPosition)
    mcontroller.setVelocity({0,0})
    return
  end

  local mechAimLimit = vehicle.configParameter("mechAimLimit") * math.pi / 180
  local mechHorizontalMovement = vehicle.configParameter("mechHorizontalMovement")
  local mechJumpVelocity = vehicle.configParameter("mechJumpVelocity")
  local mechFireCycle = vehicle.configParameter("mechFireCycle")
  local mechProjectile = vehicle.configParameter("mechProjectile")
  local mechProjectileConfig = vehicle.configParameter("mechProjectileConfig")
  local offGroundFrames = vehicle.configParameter("offGroundFrames")

  local mechCollisionPoly = mcontroller.collisionPoly()
  local position = mcontroller.position()

  if mechProjectileConfig.power then
    mechProjectileConfig.power = root.evalFunction("gunDamageLevelMultiplier", self.level) * mechProjectileConfig.power
  end

  updateVehicleDamage() -- damage visuals
  
  local entityInSeat = vehicle.entityLoungingIn("seat")
  self.driver = entityInSeat
  if not entityInSeat then
    vehicle.setDamageTeam({type = "passive"})
    animator.rotateGroup("guns", -mechAimLimit)
    animator.setAnimationState("movement", "idle")
    return
  end
  vehicle.setDamageTeam(world.entityDamageTeam(entityInSeat))
  
---- movement + animation
  local diff = world.distance(vehicle.aimPosition("seat"), mcontroller.position())
  local aimAngle = math.atan(diff[2], diff[1])
  local facingDirection = (aimAngle > math.pi / 2 or aimAngle < -math.pi / 2) and -1 or 1

  if facingDirection < 0 then
    animator.setFlipped(true)

    if aimAngle > 0 then
      aimAngle = math.max(aimAngle, math.pi - mechAimLimit)
    else
      aimAngle = math.min(aimAngle, -math.pi + mechAimLimit)
    end

    animator.rotateGroup("guns", math.pi - aimAngle)
  else
    animator.setFlipped(false)

    if aimAngle > 0 then
      aimAngle = math.min(aimAngle, mechAimLimit)
    else
      aimAngle = math.max(aimAngle, -mechAimLimit)
    end

    animator.rotateGroup("guns", aimAngle)
  end
  self.facingDirection = facingDirection
  self.aimAngle = aimAngle

  local onGround = mcontroller.onGround()
  local movingDirection = 0

  if vehicle.controlHeld("seat", "left") and onGround then
    mcontroller.setXVelocity(-mechHorizontalMovement)
    movingDirection = -1
  end

  if vehicle.controlHeld("seat", "right") and onGround then
    mcontroller.setXVelocity(mechHorizontalMovement)
    movingDirection = 1
  end

  if vehicle.controlHeld("seat", "jump") and onGround then
    -- mcontroller.setXVelocity(mechJumpVelocity[1] * movingDirection)
    -- mcontroller.setYVelocity(mechJumpVelocity[2])
    -- animator.setAnimationState("movement", "jump")
    -- self.groundFrames = 0
  end

  if onGround then
    self.groundFrames = offGroundFrames
  else
    self.groundFrames = self.groundFrames - 1
  end
  
  local emote = self.lastEmote
  local dance = self.lastDance
  local curhp = math.ceil((storage.health/self.maxHealth)*4)
  if self.groundFrames <= 0 then
    if mcontroller.velocity()[2] > 0 then
      animator.setAnimationState("movement", "jump")
    else
      animator.setAnimationState("movement", "fall")
    end
    dance = "panic" -- off ground panic
  elseif movingDirection ~= 0 then
    if facingDirection ~= movingDirection then
      animator.setAnimationState("movement", "backWalk")
    else
      animator.setAnimationState("movement", "walk")
    end
    local emotes = {"annoyed","neutral","happy","laugh"}
    dance = "typing" -- push/pull lever action
    emote = emotes[curhp]--"laugh"
  elseif onGround then
    animator.setAnimationState("movement", "idle")
    local emotes = {"sad","sad","annoyed","happy"}
    dance = "flipswitch" -- arms at sides
    emote = emotes[curhp]--"annoyed"
  end

  if emote ~= self.lastEmote then vehicle.setLoungeEmote("seat",emote) self.lastEmote = emote end
  if dance ~= self.lastDance then vehicle.setLoungeDance("seat",dance) self.lastDance = dance end
  
  if vehicle.controlHeld("seat", "primaryFire") then
-- dig fg
    bulldozeTiles("foreground")
  end
  if vehicle.controlHeld("seat", "altFire") then
-- dig bg
    bulldozeTiles("background")
  end
end

--------------------------------------------------------------------------------

function bulldozeTiles(layer)
	local bullPos = mcontroller.position()
  local bullFace = self.facingDirection
  local bullPower = vehicle.configParameter("bulldozeDamage")
  local bulldozeOffset = vehicle.configParameter("bulldozeOffset")
  local bulldozeBlocks = vehicle.configParameter("bulldozeBlocks")
  local bullYOffset = 0
  local mechAimLimit = vehicle.configParameter("mechAimLimit") * math.pi / 180

  if (self.aimAngle == mechAimLimit or self.aimAngle == (math.pi - mechAimLimit)) then bullYOffset = 1 end
  if (self.aimAngle == -mechAimLimit or self.aimAngle == (-math.pi + mechAimLimit)) then bullYOffset = -1 end
  
	bullPos = {bullPos[1], math.floor(bullPos[2])}
	
	for key,value in pairs(bulldozeBlocks) do
			bulldozeBlocks[key][1] = (value[1] + bulldozeOffset)*bullFace + bullPos[1]
			bulldozeBlocks[key][2] = value[2] + bullPos[2] + bullYOffset
      damageBlock(layer,bulldozeBlocks[key],(bullPower/script.updateDt()))
	end

end

--------------------------------------------------------------------------------
-- modified version of the tile damage code from mining gardenbot
function damageBlock(layer,block,dmg,hrv)
if dmg == nil then dmg = 1 end
if hrv == nil then hrv = 0 end
if layer == nil then layer = "foreground" end
  local dmgtiles = false
  local matname = world.material(block,layer)
  if not matname then return end
  if matname and harvestmats(matname) ~= 0 then
  hrv = 1 
  end

  local modName = world.mod(block,layer)
  if modName then --and isOre(modName) then
    dmgtiles = world.damageTiles({block},layer,block,"blockish",(0.75*dmg)) 
  else
    dmgtiles = world.damageTiles({block},layer,block,"blockish",dmg,hrv)
  end
  if not dmgtiles then -- maybe tree over ?! 
    dmgtiles = world.damageTiles({vec2.add(block,{0,1})},layer,block,"plantish",2*dmg)
  end
  if not dmgtiles then -- maybe vine under ?
    dmgtiles = world.damageTiles({vec2.add(block,{0,-1})},layer,block,"plantish",2*dmg)
  end
  if dmgtiles then
--[[
    if entity.hasSound("mine") and self.mineSoundTimer < 0 then 
      entity.playSound("mine")
      self.mineSoundTimer = 0.3 -- no earraping sounds plx
    end
--]]
  end
  util.debugRect({block[1],block[2],block[1]+1,block[2]+1},"red")
  return dmgtiles
end
--------------------------------------------------------------------------------
function harvestmats(matname)
if type(matname) ~= "string" then return 0 end -- filter out numbers and booleans
local badmats = {"rock","stone","mud","clay","sand","dirt","gravel","ice","slush","snow","moon","ash","flesh","bone","plantmat","hazard"}
for _,v in ipairs(badmats) do
if string.find(matname,v) then return 0 end
end
return 1
end

--------------------------------------------------------------------------------
-- damage to vehicle
function applyDamage(damageRequest)
  local damage = 0
  if damageRequest.damageType == "Damage" then
    damage = damage + root.evalFunction2("protection", damageRequest.damage, self.protection)
  elseif damageRequest.damageType == "IgnoresDef" then
    damage = damage + damageRequest.damage
  else
    return
  end
if damage > 0 then
  storage.health = math.max(storage.health - damage,0)
  updateVehicleHitDamage()
end

  return {{
    sourceEntityId = damageRequest.sourceEntityId,
    targetEntityId = entity.id(),
    position = mcontroller.position(),
    damage = damage,
    hitType = "Hit",
    damageSourceKind = damageRequest.damageSourceKind,
    targetMaterialKind = self.materialKind,
    killed = storage.health <= 0
  }}
end

--------------------------------------------------------------------------------
-- damage visuals
function updateVehicleDamage()  -- called per frame, randomize smoke/fire 
local curhealth = storage.health/self.maxHealth
  if curhealth > 0.8 or math.random() < curhealth then return end  -- undamaged, gtfo
  
  if curhealth <= 0.8 and math.random() > curhealth then    -- 1 smoke
    animator.burstParticleEmitter("smoke")    
  end
  if curhealth <= 0.7 and math.random() > curhealth then    -- 1 smoke
    animator.burstParticleEmitter("smoke")    
  end
  if curhealth <= 0.6 and math.random() > curhealth then
    animator.burstParticleEmitter("smoke")    
  end
  if curhealth <= 0.5 and math.random() > curhealth then
   -- 1 smoke 1 fire
    animator.burstParticleEmitter("smoke2")    
    animator.burstParticleEmitter("fire")
  end
  if curhealth <= 0.25 and math.random() > curhealth then
    -- 2 fire
    animator.burstParticleEmitter("fire")
    animator.burstParticleEmitter("fire2")
  end

end

function updateVehicleHitDamage() -- called when hit, spit out damage shards / explode when 0 hp

  animator.burstParticleEmitter("damageShards")
  if storage.health <= 0 then -- blow chunks and EXPLOSIONS!
    local projectileConfig = {
      damageTeam = { type = "indiscriminate" },
      power = vehicle.configParameter("explosionDamage"),
      onlyHitTerrain = true,
      timeToLive = 0,
      actionOnReap = {
        {
          action = "config",
          file =  vehicle.configParameter("explosionConfig")
        }
      }
    }
    world.spawnProjectile("invisibleprojectile", mcontroller.position(), 0, {0, 0}, false, projectileConfig)

    animator.burstParticleEmitter("wreckage")
    animator.burstParticleEmitter("damageShards")

    vehicle.destroy()  -- your head asplode!
  end
  
end
--------------------------------------------------------------------------------
--[[
--prints tables
function printTable(indent, value)
    local tabs = "";
    for i=1,indent,1 do
        tabs = tabs.."    ";
    end
    table.sort(value)
    for k,v in pairs(value) do
        world.logInfo(tabs..getValueOutput(k,v));
        if type(v) == "table" then
            if tostring(k) == "utf8" then
                world.logInfo("    "..tabs.."SKIPPING UTF8")-- SINCE IT SEEMS TO HAVE NO END AND JUST BE FILLED WITH TABLES OF TABLES
            else
                if tableLen(v) == 0 then
                    world.logInfo("    "..tabs.."EMPTY TABLE")
                else
                    printTable(indent+1,v);
                 
                end
            end
            world.logInfo(" ");
        end
    end
 
end

function tableLen(T)
  local count = 0
  for _ in pairs(T) do count = count + 1 end
  return count
end

--Required for printTable
function getValueOutput(key ,value)
    if type(value) == "table" then
        return "table : "..key;
    elseif type(value) == "function" then
        return "function : "..key.."()"
    elseif type(value) == "string" then
        return "string : "..key.." - \""..tostring(value).."\"";
    else
        return type(value).." : "..key.." - "..tostring(value);
    end
end
]]